function [ A ] = blkToep( a, inpStr )
%BLKTOEP Creates a block-Toeplitz matrix from the input block-vector.
%
%   T = blkToep(a) creates a block-Toeplitz matrix from the blocks in a. It
%   is assumed that the blocks are square of size m-times-m, where m is the
%   width of the matrix a. Considering a to be a block-vector, a =
%   [a_0, ..., a_k]^T, the output is the Hermitian block-Toeplitz matrix
%        ___                     ___
%       | a_0,  a_-1,  ... , a_-k   |
%   T = | a_1,  a_0,   ... , a_-k+1 |
%       |   :    :            :     |
%       | a_k, a_k-1,  ... , a_0    |
%        ---                     ---
%   where a_-l = conj(a_l)
%
%   T = blkToep(a, str) takes a string input that specify the structure of 
%   the Toeplitz matrix. If str = "LT", the function returns a lower-
%   triangular block-Toeplitz matrix
%        ___                 ___
%       | a_0,   0,   ... ,  0  |
%   T = | a_1,  a_0,  ... ,  0  |
%       |   :    :           :  |
%       | a_k, a_k-1, ... , a_0 |
%        ---                 ---
%   If str = "UT", the returned matrix is an upper-triangular block-
%   Toeplitz matrix (the block-transpose of the above matrix).

[l,k] = size(a);
n = l/k;

A = zeros(l,l);

if nargin == 1
    A(1:l,1:k) = a;
    for i = 1:n-1
        A(k+1:l, k*i+1:k*(i+1)) = A(1:l-k, k*(i-1)+1:k*i);
        A(1:k, k*i+1:k*(i+1)) = conj(a(k*i+1:k*(i+1), :));
    end
end

if nargin == 2
    if strcmpi('LT', inpStr) == 1
        A(1:l,1:k) = a;
        for i = 1:n-1
            A(k+1:l, k*i+1:k*(i+1)) = A(1:l-k, k*(i-1)+1:k*i);
        end
    elseif strcmpi('UT', inpStr) == 1
        a_trans =  blkTransp(a, k);
        A(1:k,1:l) = a_trans;
        for i = 1:n-1
            A(k*i+1:k*(i+1), k*i+1:k*n) = a_trans(:,1:k*(n-i));
        end
    else
        error('The input string should be "LT" or "UT" if lower- respective upper-triangular Toeplitz is desired, otherwise omitted')
    end
end




end

