function [ A ] = blkTransp( a, m )
%BLKTRANSP Performing block-transpose on block vectors
%
%   b = blkTransp(a, m) takes a block-row- or -column-vector a, with
%   m-times-m blocks, and returns its block-transpose. I.e. if
%        ___                 ___
%   a = | a_11, a_12, ..., a_1n |
%        ---                 ---
%   it returns
%          ______ 
%         | a_11 |
%   a^T = | a_12 |
%         |   :  |
%         | a_1n |
%          ------
%   without alterning the internal structure of the blocks, and vice versa.


if nargin ~= 2
    error('Two input arguments needed')
end

if m == 1
    A = a.';
    return
end

[l,k] = size(a);


if k < l
    % Row-vectors
    n_row = l/m;
    
    tmp1 = (1:m).';
    tmp2 = (0:m:m*(n_row-1)).';
    I = kron(ones(n_row*m,1), tmp1) + kron(ones(n_row,1), kron((tmp1-1)*n_row*m, ones(m,1))) + kron(tmp2, ones(m^2,1));

    A = reshape(a(I), m, n_row*m);
    return
    
elseif k > l
    % Column-vectors
    n_col = k/m;
    
    tmp1 = (1:m).';
    tmp2 = (0:m:m*(m-1)).';
    tmp3 = (0:n_col-1).';
    I = kron(ones(n_col*m,1), tmp1) + kron(ones(m,1), kron(tmp3*m^2, ones(m,1))) + kron(tmp2, ones(m*n_col,1));

    A = reshape(a(I), n_col*m, m);
    return
    
else
    A = a;
    return

end

%% OLD AND SLOW WAY, BUT ALSO WORKS FOR MATRICES
%
%   b = blkTransp(a, m) takes a block-matrix a, with m-times-m blocks, of
%   the form
%        ___                 ___
%       | a_11, a_12, ..., a_1n |
%   a = | a_21, a_22, ..., a_2n |
%       |   :     :          :  |
%       | a_k1, a_k2, ..., a_kn |
%   and returns the block-transpose
%          ___                 ___
%         | a_11, a_21, ..., a_k1 |
%   a^T = | a_12, a_22, ..., a_k2 |
%         |   :     :          :  |
%         | a_1n, a_2n, ..., a_kn |
%   without alterning the internal structure of the blocks.
%
% n_col = k/m;
% n_row = l/m;
% 
% 
% A = a.';
% 
% for i = 1:n_row
%     for j = 1:n_col
%         % The transposed matrix has n_row columns and n_col rows
%         A(m*(j-1)+1:m*j,m*(i-1)+1:m*i) = A(m*(j-1)+1:m*j,m*(i-1)+1:m*i).';
%     end
% end


end

