function [ output ] = fastToepMult( matrix, vector, inpStr )
%FASTTOEPMULT Computes block-matrix - block-vector multiplication A*b,
%   where A is a block-Toeplitz matrices, exploting the structure of A to
%   perform it in O(2N log(2N)) instead of O(N)
%
%   x = fastToepMult(A, b) returns the result of multiplying the
%   Toeplitz matrix A, which is represented by the first column in A, with
%   the vector b. The matrix is assumed to be Hermitian symmetric.
%
%   x = fastToepMult(A, b, str) also returns the multiplication, but str
%   takes as input 'LT' or 'UT', in order to specify if the matrix is
%   lower- respective upper-triangular.
%
%   The computations are done by extending the representation A into a
%   representation of a circulant matrix with double number of rows and
%   columns. By extending the vector b in the same way, as
%       tilde_b = [b; 0]
%   the first element of the multiplication with the circulant matrix will
%   be the sought answer. Now, circulant matrices are diagonalized by FFT
%   and hence the multiplication can be done by performing one set FFTs,
%   then pointwise multiplication of the elements, and finally performing a
%   second set of inverse FFTs.
%

if nargin < 2
    error('Need at least two input argument')
end

if nargin > 3
    error('Allows maximum three input arguments')
end


[l,k] = size(matrix);
n = l/k;

[l_v,k_v] = size(vector);


%% Creating the circulant represenation
if nargin == 3
    % In this case the Toeplitz matrix is triangular
    if strcmpi('LT', inpStr) == 1
        Z = zeros(l,k);
        fft_mtx = [matrix;Z];
    
    elseif strcmpi('UT', inpStr) == 1
        d_rev = blkFlipud(matrix, k);
        Z = [zeros(k,k); d_rev(1:k*(n-1),:)]; 
        fft_mtx = [matrix(1:k,:); zeros(k*(n-1), k); Z];
    
    else
        error(strcat('The input string should be "LT" or "UT" if the multiplication is to be done with',...
            ' lower- respective upper-triangular Toeplitz, otherwise omitted'))
    end
    
elseif nargin == 2
    % Here it is assumed that the matrix is block-Hermitian symmetric
    d_rev = blkFlipud(matrix, k);
    Z = conj([matrix(1:k,:); d_rev(1:k*(n-1),:)]);
    fft_mtx = [matrix; Z];

else
    % Cannot get here
    error('Need one or two input arguments')
end

%% Computing the FFTs, which is done 'block-wise'
%(i.e. all elements in the same block should be multiplied with the same frequencey)
fft_vec = [vector;zeros(l,k_v)];

mtx_rep = zeros(2*l,k);
vec_rep = zeros(2*l,k_v);

for i = 1:k
    for j = 1:k
        mtx_rep(i:k:end, j) = fft(fft_mtx(i:k:end, j));
%         vec_rep(i:k:end, j) = fft(fft_vec(i:k:end, j));
    end
    for j = 1:k_v
        vec_rep(i:k:end, j) = fft(fft_vec(i:k:end, j));
    end
end

% % This implementation does the same as the for-loops above, however it only
% % works for the case when the matrix and the vector have the same
% % block-size, and not for "scalar"-vectors. Moreover, it seems slower (at
% % least for small blocks).
% 
% vec_rep = fft_vec.';
% vec_rep = fft(reshape(vec_rep(:), k_v^2, l).');
% vec_rep = reshape(vec_rep.', k_v, k_v*l).';
% 
% mtx_rep = fft_mtx.';
% mtx_rep = fft(reshape(mtx_rep(:), k_v^2, l).');
% mtx_rep = reshape(mtx_rep.', k_v, k_v*l).';
% 
% % Parts of an even older version of the above
% %vec_rep = blkTransp(fft_vec, k_v);
% %mtx_rep = blkTransp(fft_mtx, k_v);
% % I = kron(ones(k,1), kron((1:2*n).', ones(k,1)) + kron(ones(2*n,1), (0:2*n:2*n*(k-1)).') ) + kron((0:k - 1).'*2*n*k, ones(2*n*k,1));
% % 
% % vec_rep = reshape(vec_rep(I), k_v*l, k_v);
% % mtx_rep = reshape(mtx_rep(I), k*l, k);


%% Performing multiplication and block-wise IFFT

multipResult = sprBlkDiag(mtx_rep)*vec_rep;

% % Does the same thing as the line above, but with a for-loop. The line
% % above utilizes that Matlab is (supposingly) optimized for operations
% % with sparse matrices.
% multipResult = zeros(2*l,k_v);
% 
% for i = 1:2*n
%     multipResult(k*(i-1)+1:k*i,:) = mtx_rep(k*(i-1)+1:k*i,:)*vec_rep(k*(i-1)+1:k*i,:);
% end


product = zeros(2*l,k_v);
for i = 1:k
    for j = 1:k_v
        product(i:k:end, j) = ifft(multipResult(i:k:end, j));
    end
end


output = product(1:l,:);


end

