function [ x ] = fundSol( a, b )
%FUNDSOL Computes the first set of fundamental solutions Ax_i = g_i, i = 1,
%   2, 3 and 4, where A > 0 is a block-Toeplitz-plus-block-Hankel matrix
%
%   x = fundSol(a,b) computes the set of fundamental solutions Ax_i = g_i,
%   where A is a positive definite block-Toeplitz-plus-block-Hankel matrix,
%   i.e. A = [a_{i-j} + b_{i+j}]. The input a and b are two column vectors
%   where a is to the first column of the Toeplitz matrix [a_0, ..., a_n]^T
%   and b contains both the first column and the last row of the Hankel
%   matrix [b_0, ..., b_n, ..., b_2n]^T. The entries a_i, b_i can be scalar
%   or matrix valued, with real or compex entries. x is a (n+1)x4k matrix,
%   where each block i of k columns, i.e. each block x(:, k*(i-1) + 1:k*i),
%   is the fundamental solution x_i
%
%   The algorithm generalizes the one, version AIII, presented in
%   Heinig, Jankowski, Rost. Fast inversion of Toeplitz-plus-Hankel
%   Matrices. 1988.

%% Checking indata
if nargin ~= 2
    error('Need exactly two input arguments')
end

[l,k] = size(a);
if rem(l,k) ~= 0
    error('The input vector a does not consist of an integer number of blocks')
end

[l_b, k_b] = size(b);
if rem(l_b,k_b) ~= 0
    error('The input vector b does not consist of an integer number of blocks')
end

if k ~= k_b
    error('The block-size of a and b must be the same')
end

% Number of blocks in a and b
n = l/k;
n_b = l_b/k_b;
if 2*n-1 ~= n_b
    error('The number of blocks in a and b are not consistent with each other')
end


%%%%%%%%
% TODO %
%%%%%%%%
% Should also check the positive definitness of A = [a_{i-j} + b{i+j}], but
% I don't want to form the matrix explicitly... Can this been seen
% somewhere in the recursion?



%% Setting up variables and initial calculations 

% Defining the "auxilary" values to 0
a = [a; zeros(k)];
b = [zeros(k); b; zeros(k)];

% Creating the reversed vector a_rev = [a_n, ..., a_0]^T
a_rev = blkFlipud(a, k);
a_rev_bt = blkTransp(a_rev, k);


% Forming the recursive vectors among the g:s and f:s. Note that the f:s
% are defined as block-row-vectors in order to not have to block-transpose
% them in the loop.
g1 = [eye(k); zeros(k*(n-1),k)];
g2 = [zeros(k*(n-1),k); eye(k)];
g3 = - ( a(k+1:(n+1)*k, :) + b(1:n*k, :) );

a_bt = blkTransp(a, k);
b_bt = blkTransp(b, k);

f1 = conj( a_bt(:, k+1:(n+1)*k) ) + b_bt(:, 1:n*k);
f3 = blkTransp(g1, k);


% Variable init.
x1 = zeros(k*n,k);
x2 = zeros(k*n,k);
x3 = zeros(k*n,k);
x4 = zeros(k*n,k);

% % Not needed in this formulation, where the shift-up + shift-down was
% % performed explicitly
% % Forming the matrix U = shift-up + shift-down matrix
% S = spdiags(ones(k*(n-1)), -k, n*k, n*k);
% U = S + S.';


% Compute x4^1 and xi^2 for i = 1,2,3,4 by solving the explicit equations.
% xi^2, for i = 1,2,3,4, are computed to initialize the recursion, while
% x4^1 is needed to compute alpha^1, which is needed to compute lambda^2,
% which in turn are needed to compute x4^3 in the loop. x4^1 is also needed
% explicitly in these computation.

A_1 = a(1:k, :) + b(k+1:2*k, :);

g4_1 = - (conj(a(k+1:2*k, :)) + b(2*k+1:3*k, :));
x4_1 = A_1\g4_1;

f2_1 = a_bt(:, k+1:2*k) + b_bt(:, 2*k+1:3*k);

alpha_1 = f2_1 * x4_1 + a(1:k, :) + b(3*k+1:4*k, :);



A_2 = [a(1:k, :) + b(k+1:2*k, :), conj(a(k+1:2*k, :)) + b(2*k+1:3*k, :)
    a(k+1:2*k, :) + b(2*k+1:3*k, :), a(1:k, :) + b(3*k+1:4*k, :)];

x1(1:2*k, :) = A_2\g1(1:2*k, :);
x2(1:2*k, :) = A_2\g2(end-2*k + 1:end, :);
x3(1:2*k, :) = A_2\g3(1:2*k, :);

g4_2 = -( conj( a_rev(end-3*k+1:end-k, :) ) + b(3*k+1:5*k, :) );
x4(1:2*k, :) = A_2\g4_2;

f2_2 = a_rev_bt(:, end-3*k+1:end-k) + b_bt(:, 3*k+1:5*k);


% Need x4_m-1 in order to calculate x4_m+1 and y2_m+1, therefore store the
% solution for m = 1 in the other varaible
x4_mM1 = [x4_1; zeros(k*(n-1),k)];

% Store the inverse of alpha^1 as the inverse is caluculated in previous
% iteration in the loop
alpha_m_inv = inv(alpha_1);

% Calculate lambda^2. For m = 2, both ^2 and ^3 are needed, hence these
% will later in the loop be stored in the variables lambda_m
lambda_mP1 = f2_2 * [x4_1(1:k, :); eye(k)];


%% The recursion

% Running the loop from 2 to n-1, and in each iterartion computing the
% solution m+1.

for m = 2:n-1
    %% x-part of the calculations
    
    f2_m = a_rev_bt(:, end-(m+1)*k + 1:end-k) + b_bt(:, (m+1)*k + 1: (2*m+1)*k); 
    alpha_mM1_inv = alpha_m_inv;
    
    alpha_m = f2_m* x4(1:m*k, :) + a(1:k, :) + b((2*m+1)*k+1: (2*m+2)*k, :);    
    alpha_m_inv = inv(alpha_m);
    
    gamma_1m = f2_m * x1(1:m*k, :) - g1(m*k+1:(m+1)*k, :);
    gamma_3m = f2_m * x3(1:m*k, :) - g3(m*k+1:(m+1)*k, :);
    
    x2(1:(m+1)*k, :) = [x4(1:m*k, :); eye(k)] * alpha_m_inv;
    x1(1:(m+1)*k, :) = [x1(1:m*k, :); zeros(k)] - x2(1:(m+1)*k, :) * gamma_1m;
    x3(1:(m+1)*k, :) = [x3(1:m*k, :); zeros(k)] - x2(1:(m+1)*k, :) * gamma_3m;

    
        
    f2_mP1 = a_rev_bt(:, end-(m+2)*k + 1:end-k) + b_bt(:, (m+2)*k + 1: (2*m+3)*k); 
    
    lambda_m = lambda_mP1;
    lambda_mP1 = f2_mP1 * [x4(1:m*k, :); eye(k)];
    
    beta_1m = f1(:, 1:(m+1)*k) * [x4(1:m*k, :); eye(k)];
    beta_3m = f3(:, 1:(m+1)*k) * [x4(1:m*k, :); eye(k)];
    
    tmp_x4 = x4;    

%     % x4-update done with explicit multiplication with the shift matrix  
%     x4(1:(m+1)*k, :) = ...
%         U(1:(m+1)*k,1:(m+1)*k) * [x4(1:m*k, :); eye(k)]...
%         - [x4_mM1(1:(m-1)*k, :); eye(k); zeros(k)] * alpha_mM1_inv * alpha_m ...
%         + [x4(1:m*k, :); eye(k)] * (alpha_m_inv * (lambda_m * alpha_mM1_inv * alpha_m - lambda_mP1)) ...
%         - x1(1:(m+1)*k, :)*beta_1m - x3(1:(m+1)*k, :)*beta_3m;
    
    x4(1:(m+1)*k, :) = ...
        [zeros(k); x4(1:m*k, :)] + [x4(k+1:m*k, :); eye(k); zeros(k)]...
        - [x4_mM1(1:(m-1)*k, :); eye(k); zeros(k)] * alpha_mM1_inv * alpha_m ...
        + [x4(1:m*k, :); eye(k)] * (alpha_m_inv * (lambda_m * alpha_mM1_inv * alpha_m - lambda_mP1)) ...
        - x1(1:(m+1)*k, :)*beta_1m - x3(1:(m+1)*k, :)*beta_3m;

    x4_mM1 = tmp_x4;
       
end

x = [x1, x2, x3, x4];



end

