function [ solVec ] = thSolver( t, h, vec )
%THSOLVER Solves the equation Ax = b, where A is a block-Toeplitz plus
%   block-Hankel matrix. If the block-size is k, and the number of blocks
%   are n, the complexity is O(kn^2) + O(kn log(n)).
%
%   x = thSolver(t, h, b) solves the system (T + H)x = b, where T is a
%   block-Toeplitz and H is a block-Hankel matrix. t and h are block-column
%   vectors containing a representation of the block-Toeplitz respective
%   block-Hankel matrices, as shown below
%          ___                 ___                     ___ 
%        | t_0 |              | t_0,  t_-1,  ... , t_-n   | 
%   t =  |  :  |     =>   T = | t_1,  t_0,   ... , t_-n+1 |
%        | t_n |              |   :    :            :     |
%          ---                | t_n,  t_n-1, ... , t_0    |
%                              ---                     --- ,
%   where t_-l = conj(t_l),
%          _____                 ___                  ___ 
%        |  h_0  |              | h_0,  h_1,  ... ,  h_n   | 
%        |   :   |     =>   H = | h_1,  h_2,   ... , h_n+1 |
%   h =  |  h_n  |              |   :    :            :    |
%        | h_n+1 |              | h_n,  h_n+1, ... , h_2n  |
%        |   :   |               ---                    --- .
%        | h_2n  |
%          -----
%
%   The solution is computed using the functions fundSol and fastToepMult.

%% Check input
if nargin ~= 3
    error('Need exactly three input arguments; the Toeplitz and the Hankel vectors defining the matrix, and the vector to solve for')
end

[l,k] = size(t);
if rem(l,k) ~= 0
    error('The input vector for the Toeplitz part does not consist of an integer number of blocks')
end

[l_b, k_b] = size(h);
if rem(l_b,k_b) ~= 0
    error('The input vector for the Hankel part does not consist of an integer number of blocks')
end

% Block-size of a and b
if k ~= k_b
    error('The block-size of the Toeplitz and the Hankel part must be the same')
end


% Number of blocks in a and b
n = l/k;
n_b = l_b/k_b;
if 2*n-1 ~= n_b
    error('The number of blocks in  the Toeplitz and the Hankel part are not consistent with each other')
end


[l_v, k_v] = size(vec);

if k_v ~= 1
    if k_v ~= k
        error('The block-size of the vector need to be 1 or the same as the matrix')
    end
end

if l ~= l_v
    error('The number of rows in the matrix and the vector are not consistent with each other')
end    
    

%% Solve the system

% Computing the fundamental solutions
%[x, y] = fundSol_xy(t,h);
x = fundSol(t,h);
y = [blkTransp(-x(:,2*k+1:3*k).',k), blkTransp(-x(:,3*k+1:4*k).',k), blkTransp(x(:,1:k).',k), blkTransp(x(:,k+1:2*k).',k)];

% Forming the vectors b and c of the coefficients for the generating
% polynomials of the inverse
b = zeros(k*(n+2),4*k);
c = zeros(k*(n+2),4*k);

b(:, 1:k) = [zeros(k); x(:,1:k); zeros(k)];
b(:, k+1:2*k) = [zeros(k); x(:,k+1:2*k); zeros(k)];
b(:, 2*k+1:3*k) = [eye(k); x(:,2*k+1:3*k); zeros(k)];
b(:, 3*k+1:4*k) = [zeros(k); x(:,3*k+1:4*k); eye(k)];

c(:,1:k) = [-eye(k); y(:, 1:k); zeros(k)];
c(:, k+1:2*k) = [zeros(k); y(:, k+1:2*k); -eye(k)];
c(:, 2*k+1:3*k) = [zeros(k); y(:, 2*k+1:3*k); zeros(k)];
c(:, 3*k+1:4*k) = [zeros(k); y(:, 3*k+1:4*k); zeros(k)];


% Forming the other quantities needed to multiply with the inverse
e = kron(ones(floor(n/2),1), [eye(k);zeros(k)]);
if mod(n,2) == 1
    e = [e;eye(k)];
end

R = kron(ones(floor(n/2),1), [zeros(k);eye(k)]);
if mod(n,2) == 1
    R = [R;zeros(k)];
end

hat_c = zeros(k*(n+2),4*k);
for i = 1:4
    hat_c(:,k*(i-1)+1:k*i) = blkFlipud(c(:,k*(i-1)+1:k*i),k);
end


% Performing the implicit multiplications
solVec = zeros(l,k_v);

for i = 1:4
    solVec = solVec + ...
        fastToepMult(...
            b(1:k*n,k*(i-1)+1:k*i), ...
                fastToepMult(R, ...
                    blkFlipud( fastToepMult(hat_c(1:k*n,k*(i-1)+1:k*i), vec, 'LT'), k) ...
                    + fastToepMult(c(1:k*n,k*(i-1)+1:k*i), vec, 'UT'), ...
                'LT')...
                + ( e * blkTransp( c(k+1:k*(n+1),k*(i-1)+1:k*i), k) * vec ), ...
            'LT' ...
        );
end

end


