function [ q, p, newtonDecrements, numIter] = newtonSolver_crce_fast(c, m, lambda, N)
% Solves the problem regularized CRCE-problem with cepstral matching, by
% solving the potimization problem
%       min c^T x - m^T y
%            - sum_{j=1}^{N} [Fy]_j log([Fy]_j / [Fx]_j)
%            - lambda sum_{j=1}^{N} log([Fy]_j)
%
% The info struct contains entries exitFlag and numericalPorblems.
%
% exitFlag = 1 means everything is fine, exitFlag = 0
% means in-data was wrong, exitFlag = -1 means that the problem is
% unbounded, exitFlag = -2 means that c was indicated to not be in the dual
% cone during data-checking but no unbouded direction was found during 500,
% exitFlag = -3 means an unbounded direction was found but the current
% point is not feasible (check the numericalProblem if = 1).
% 
% numericalProblems = 0 means everything is fine, numericalProblems = 1
% means that the Hessian is not positive definite (accorind to Matlabs
% chol) or the Newton decrement has become negative.




%% Extract info
n = length(c);

%% Algorithm parameters and output data
%Paramters
alpha = 1e-1;
beta = 0.9;
gamma = 0.5;
iterLimit = 300;


% %%%% TESTING WITH ME AS INITIAL GUESS %%%%%
% % Does not seem to earn anything by starting at the ME-solution
% % Fewer iterations, but all the saved time is lost in computing the
% % solution
% 
% q = levinson(c,n-1);
% 
% % Computing spectrum and covariances
% q_ME = conv(fliplr([1/2*q(1), q(2:end)]), [1/2*q(1), q(2:end)]);
% q_ME = q_ME(n:end);
% q_ME = q_ME';
% Q_ME = real(fft([q_ME(1); 1/2*q_ME(2:end); zeros(N-2*n+1, 1); 1/2*flipud(q_ME(2:end))]));
% % q_spectrum_ME = flipud(q_ME);
% % denominator_ME = @(z) abs(polyval(q_spectrum_ME, exp(-1i*z))).^2;
% % Q_ME = denominator_ME(F);
% c_est_ME = ifft(1./Q_ME);
% 
% % Normalizing
% q = q * 1/sqrt(c(1)/c_est_ME(1));
% q = conv(fliplr([1/2*q(1), q(2:end)]), [1/2*q(1), q(2:end)]);
% q = q(n:end);
% q = q';
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


q = [1;zeros(n-1,1)];
p = [1;zeros(n-1,1)];


%%%% Speed improvement %%%%
% To improve speed, remove the computation of the Newton decrement
%%%%%%%%%%%%%%%%%%%%%%%%%%%
newtonDecrements = zeros(1, iterLimit);


%% Check in-data
% %%%% Speed improvement %%%%
% % To improve speed, these checks have been removed
% %%%%%%%%%%%%%%%%%%%%%%%%%%%
% if isempty( find( F(:,1) < 0) ) ~= 1
%     disp('Solver need F with first column positive')
%     info.exitFlag = 0;
%     return
% end
% 
% F = dftmtx(N);
% F = [F(:,1), real( F(:,2:n+1) )];
% [~, checkOnC, isUnbounded] = linprog(c,-F,zeros(N,1));
% if checkOnC < -1e-5 && isUnbounded == 1
%     error('c is not in the dual cone to P')
% elseif isUnbounded == -3
%     disp('c seems not to be in the dual cone to P.')
%     disp('However, since this is not sure the optimization is still run')
%     disp('but the results must be check carefully.')
% end


%% Newton iteartions 
%Initializing variables
P = real(fft([p(1); 1/2*p(2:end); zeros(N-2*n+1, 1); 1/2*flipud(p(2:end))]));
Q = real(fft([q(1); 1/2*q(2:end); zeros(N-2*n+1, 1); 1/2*flipud(q(2:end))]));

c_est = ifft(P./Q);
c_est = real(c_est(1:n));
m_est = ifft( log( P ./ Q ) - lambda* 1 ./ P );
m_est = real(m_est(2:n));

iterCount = 0;


while norm([(c-c_est)./c; (m-m_est)./m], 2) > 1e-3 && iterCount < iterLimit
    iterCount = iterCount + 1;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Profiler estimates 25% of the time in this function is spent here...
    % Remove this line for speed benifits
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     disp(strcat('Iteration ', [' ', num2str(iterCount)] ));
    
   
    gradient_1 = c - c_est;
    gradient_2 = -m + m_est;

    % thSolver
    gradient = zeros(2*n-1,1);
    gradient(1) = gradient_1(1);
    gradient(2:2:end) = gradient_1(2:end);
    gradient(3:2:end) = gradient_2;

%     % \-solver
%     gradient_debugg = [gradient_1; gradient_2];
    
    
    hessian_11 = 1/2*ifft(P ./ Q.^2);
    hessian_12 = -1/2*ifft(1./Q);
    hessian_22 = 1/2*ifft(1 ./ P + lambda * 1 ./ P.^2);
    
    
    % thSolver
    d0 = 2*hessian_11(1,1);
    
    d = zeros(1,2*(n-1));
    d(1:2:end) = 2*hessian_11(2:n).';
    d(2:2:end) = 2*hessian_12(2:n).';
    
    D = zeros(2*(2*(n-1) + 1),2);
    D(1:2:end,1) = hessian_11(1:2*n-1);
    D(1:2:end,2) = hessian_12(1:2*n-1);
    D(2:2:end,1) = hessian_12(1:2*n-1);
    D(2:2:end,2) = hessian_22(1:2*n-1);
    
    T = D(1:2*(n-1),:);
    H = D(5:end,:);
    
    %tmp1 = thSolver(T,H,d.');
    %tmp2 = thSolver(T,H,gradient(2:end));
    
    tmp=thSolver(T,H,[d.', gradient(2:end)]);
    tmp1=tmp(:,1);
    tmp2=tmp(:,2);
    
    searchDirection_1 = 1/(d0 - d*tmp1)*( gradient(1) - d*tmp2 );
    searchDirection_2 = tmp2 - tmp1*searchDirection_1;
    
    searchDirection = -[searchDirection_1; searchDirection_2];
    
    
%     % \-solver
%     hessian = [d0, d; d.', blkToep(T) + blkHank(H)];
%     searchDirection_test = -hessian\gradient;
    
    
    delta_q = [searchDirection(1); searchDirection(2:2:end)];
    delta_p = searchDirection(3:2:end);
    

%     newtonDecrement = real(searchDirection' * hessian * searchDirection / 2);
%     newtonDecrements(iterCount) = newtonDecrement;

    
    Z_q = real(fft([delta_q(1); 1/2*delta_q(2:end); zeros(N-2*n+1, 1); 1/2*flipud(delta_q(2:end))]));
    z_qLessThanZero = find(Z_q < 0);
    
    Z_p = real(fft([0; 1/2*delta_p; zeros(N-2*n+1, 1); 1/2*flipud(delta_p)]));
    Z_p = Z_p(2:end);
    z_pLessThanZero = find(Z_p < 0);

    
% %%%% Speed improvement %%%%
% % To improve speed, these checks have been removed
% %%%%%%%%%%%%%%%%%%%%%%%%%%%   
%     % Checking for numerical problems
%     [~, pdIndex] = chol(hessian);
%     if pdIndex ~= 0 || newtonDecrement < 0
%         error('The Hessian is not pdf anymore, which should be impossible.')
%     end
%     
% 
%     % Checking for unboundedness
%     tmp = c'*searchDirection(1:n);
%     if tmp < 0 && isempty(z_qLessThanZero) == 1
%         if isempty(find(F*q < 0)) == 1
%             error('The problem is unbounded.')
%         else
%             disp('Unbounded direction found, but current point is not feasible. Indicates numerical errors.')
%         end
%     end


    % If there are any indicies less than zero, compute maximum step length
    % Do it for both p and q
    if isempty(z_qLessThanZero) == 0
        t_max_q = min( abs( Q(z_qLessThanZero)./Z_q(z_qLessThanZero) ) );
    else
        %Just making sure gamma*t_max > 1, since no index put any
        %constraint on the step length
        t_max_q = 2/gamma;
    end
    
    if isempty(z_pLessThanZero) == 0
        t_max_p = min( abs( P(1 + z_pLessThanZero)./Z_p(z_pLessThanZero) ) );
    else
        %Just making sure gamma*t_max > 1, since no index put any
        %constraint on the step length
        t_max_p = 2/gamma;
    end
    
    % Take the shortest of the two, or take 1 as initial step length
    if t_max_p < 1 || t_max_q < 1
        stepLength = min([gamma*t_max_q, gamma*t_max_p,1]);
    else
        stepLength = 1;
    end
    
    
    q_new = q + stepLength * delta_q;
    p_new = [1; p(2:end) + stepLength * delta_p];
    
    P_new = real(fft([p_new(1); 1/2*p_new(2:end); zeros(N-2*n+1, 1); 1/2*flipud(p_new(2:end))]));
    Q_new = real(fft([q_new(1); 1/2*q_new(2:end); zeros(N-2*n+1, 1); 1/2*flipud(q_new(2:end))]));   

    stepLenghtIterIndex = 0;
    minDecreaseValue = c' * q - m' * p(2:end) + 1/N*P' * log( P ./ Q ) - 1/N*lambda * sum(log(P));
    
    % This is the so called Armijo-condition, for not to long step
    while c' * q_new - m' * p_new(2:end) + 1/N*P_new' * log( P_new ./ Q_new ) - 1/N*lambda * sum(log(P_new)) ...
            > minDecreaseValue + alpha * stepLength * gradient' * searchDirection... 
            && stepLenghtIterIndex < 100
        stepLength = beta * stepLength;
        q_new = q + stepLength * delta_q;
        p_new = [1; p(2:end) + stepLength * delta_p];
        P_new = real(fft([p_new(1); 1/2*p_new(2:end); zeros(N-2*n+1, 1); 1/2*flipud(p_new(2:end))]));
        Q_new = real(fft([q_new(1); 1/2*q_new(2:end); zeros(N-2*n+1, 1); 1/2*flipud(q_new(2:end))])); 
        stepLenghtIterIndex = stepLenghtIterIndex + 1;
    end
    
    q = q_new;
    p = p_new;
    
    P = P_new;
    Q = Q_new;
    
    c_est = ifft(P ./ Q);
    c_est = real(c_est(1:n));
    m_est = ifft( log(P ./ Q) - lambda* 1 ./ P);
    m_est = real(m_est(2:n));
end

newtonDecrements = newtonDecrements(1:iterCount);
numIter = iterCount;

end